/* Implementation of TLInetAddress class.
   This file is part of TL, Tiggr's Library.
   Written by Tiggr <tiggr@es.ele.tue.nl>
   Copyright (C) 1995, 1996 Pieter J. Schoenmakers
   TL is distributed WITHOUT ANY WARRANTY.
   See the file LICENSE in the TL distribution for details.

   $Id: TLInetAddress.m,v 1.13 1996/01/23 12:18:32 tiggr Exp $  */

#define TLINETADDRESS_DECLARE_PRIVATE_METHODS
#import "tl/TLInetAddress.h"
#import "tl/TLInetSocketStream.h"
#import "tl/TLPatchedRoots.h"
#import "tl/TLString.h"
#import "tl/TLNumber.h"
#import "tl/support.h"
#import <netdb.h>
#import <sys/socket.h>

@implementation TLInetAddress

+(TLInetAddress *) addressWithProtocol: (id <TLString>) pr
 port: (id <TLString>) portname host: (id <TLString>) hostname
{
  return ([[self gcAlloc] initWithProtocol: pr host: hostname port: portname]);
} /* addressWithStream:host:port: */

+(id <TLNetworkAddress>) addressWithAddress: (void *) addr length: (int) len
{
  return ([[self gcAlloc] initWithAddress: addr length: len]);
} /* +addressWithAddress:length: */

+(id <TLNumber>) portFromService: (id <TLString>) name
			protocol: (id <TLString>) protoc0l
{
  struct servent *sent;

  sent = getservbyname ((void *) [name cString], (void *) [protoc0l cString]);
  return sent ? [TLNumber numberWithInt: sent->s_port] : nil;
} /* +portFromService:protocol: */

-(void *) address
{
  return (&address);
} /* -address */

-(int) addressSize
{
  return (sizeof (address));
} /* -addressSize */

-(int) compare: (id <TLInetAddress>) a
{
  /* XXX Should compare addressen?  TLHost should be involved with
     multiple hostnames and multiple addresses.  */
  return ([host compare: [a hostname]]
	  || [port compare: [a port]]
	  || [protocol compare: [a protocol]]);
} /* -compare: */

-(id <TLString>) hostname
{
  if (!host)
    {
      struct hostent *e = gethostbyaddr ((void *) &address.sin_addr,
					 sizeof (address.sin_addr), AF_INET);
      if (e)
	host = [TLString stringWithCString: e->h_name];
      else
	host = [TLString stringWithFormat: @"%d.%d.%d.%d",
			 (address.sin_addr.s_addr >> 24) & 0xff,
			 (address.sin_addr.s_addr >> 16) & 0xff,
			 (address.sin_addr.s_addr>> 8) & 0xff,
			 address.sin_addr.s_addr & 0xff];
    }
  return (host);
} /* -hostname */

-initWithProtocol: (id <TLString>) proto
 host: (id <TLString>) hostname port: (id) portname
{
  const char *pr = [proto cString];
  struct servent *sent;
  struct hostent *hent;
  const char *h;

  ASGN_IVAR (host, hostname);
  ASGN_IVAR (port, portname);
  ASGN_IVAR (protocol, proto);

  /* Is this a name or a number?  */
  if (portname && [portname conformsTo: @protocol (TLString)]
      /* Cast to void * to get rid of const qualification for systems where
	 getservbyname accepts plain char *'s (NeXT).  */
      && ((sent = getservbyname ((void *) [portname cString], (void *) pr))))
    address.sin_port = sent->s_port;
  else if (portname && [portname conformsTo: @protocol (TLNumber)])
    address.sin_port = [portname integerPIntValue];
  else
    {
      [self warning: "Unknown service or port: `%#/%s'", portname, pr];
      INIT_RETURN_NIL ();
    }

  if (hostname)
    {
      h = [hostname cString];
      hent = gethostbyname ((void *) h);
      if (!hent)
	{
	  /* XXX Should report with ERRNO or H_ERRNO (hpux).  */
	  [self warning: "Unknown host: `%s'", h];
	  INIT_RETURN_NIL ();
	}    
      address.sin_family = hent->h_addrtype;
      /* XXX Should handle the list of addresses...  */
      memcpy (&address.sin_addr, (void *) hent->h_addr_list[0],
	      hent->h_length);
    }
  else
    {
      address.sin_family = AF_INET;
      address.sin_addr.s_addr = 0;
    }

  return (self);
} /* -initWithProtocol:host:port: */

-initWithAddress: (void *) addr length: (int) len
{
  /* This is an insane sanity check, but it works.  */
  if (len != sizeof (address))
    {
      [self warning: "Bad address length %d != %d", len, sizeof (address)];
      INIT_RETURN_NIL ();
    }

  memcpy (&address, addr, len);

  /* XXX Set host, port and the like now, or later?  */
  return (self);
} /* -initWithAddress:length: */

-(id <TLString>) port
{
  return (port);
} /* -port */

-(void) print: (id <TLOutputStream>) stream quoted: (BOOL) qp
{
  int named = !qp;

  if (qp)
    formac (stream, @"(%s ", [isa name]);

  if (named)
    {
      id <TLString> h = [self hostname];

      if (h)
	formac (stream, @"%@:%u", h, address.sin_port);
      else
	named = 0;
    }

  if (!named)
    formac (stream, @"%u.%u.%u.%u:%u", address.sin_addr.s_addr >> 24,
	    (address.sin_addr.s_addr >> 16) & 0xff,
	    (address.sin_addr.s_addr >> 8) & 0xff,
	    address.sin_addr.s_addr & 0xff, address.sin_port);

  if (qp)
    formac (stream, @")");
} /* -print:quoted: */

-(id <TLString>) protocol
{
  return (protocol);
} /* -protocol */

-(id) socketClass
{
  return ([TLInetSocketStream self]);
} /* -socketClass */

/******************** garbage collection ********************/

-(void) gcReference
{
  MARK (host);
  MARK (port);
  MARK (protocol);
} /* -gcReference */

@end
