/* Implementation of the TLURL class.
   This file is part of TL, Tiggr's Library.
   Written by Tiggr <tiggr@es.ele.tue.nl>
   Copyright (C) 1995, 1996 Pieter J. Schoenmakers
   TL is distributed WITHOUT ANY WARRANTY.
   See the file LICENSE in the TL distribution for details.

   $Id: TLURL.m,v 1.11 1996/01/23 12:18:43 tiggr Exp $  */

#define TLURL_DECLARE_PRIVATE_METHODS

#import "tl/TLURL.h"
#import "tl/support.h"
#import "tl/TLInetAddress.h"
#import "tl/TLPatchedRoots.h"
#import "tl/TLMutableString.h"
#import "tl/TLCons.h"
#import <stdlib.h>

#define SEPARATOR(X)  \
  ((X) == ':' || (X) == '/' || (X) == ';' || (X) == '?' || (X) == '#')

/* Same as SEPARATOR, but also allow `:', since that has lost its special
   meaning this far in the URI.  */
#define SEPARATOR2(X)  \
  ((X) == '/' || (X) == ';' || (X) == '?' || (X) == '#')

/* Same as SEPARATOR2, but also allow `/', since that has lost its special
   meaning this far in the URI.  */
#define SEPARATOR3(X)  \
  ((X) == ';' || (X) == '?' || (X) == '#')

/* Sensible URI definition:

   uri		: [scheme] server [abs_path] [params] [query] [fragment]
		| [scheme] [path] [params] [query] [fragment]

   scheme	: .+ `:'

   server	: `//' server_name [ `:' port ]

   path		: rel_path
		| abs_path

   abs_path	: `/' [rel_path]

   params	: ( `;' param )+

   query	: `?' .+ [arglist]
		| `?' args

   args		: arg arglist

   arglist	: `&' args

   fragment	: `#' .+

   */

@implementation TLURL

+(id <TLURL>) urlWithString: (id <TLString>) name
{
  return ([[self gcAlloc] initWithString: name relativeTo: nil]);
} /* +urlWithString: */

+(id <TLURL>) urlWithString: (id <TLString>) name
		 relativeTo: (id <TLURL>) url
{
  return ([[self gcAlloc] initWithString: name relativeTo: url]);
} /* +urlWithString:relativeTo: */

+(id <TLURL>) urlWithURL: (id <TLURL>) url
	      relativeTo: (id <TLURL>) base
{
  return ([[self gcAlloc] initWithURL: url relativeTo: base]);
} /* +urlWithURL:relativeTo: */

-(void) addPathElement: (id) e
{
  [path addElement: e];
  filename = nil;
} /* -addPathElement: */

-(id <TLString>) basename
{
  int l = [path length];
  return l ? [path _elementAtIndex: l - 1] : nil;
} /* -basename */

-(int) compare: (id <TLURL>) url
{
  int i = 0;
  id x;

#define COMPARE(IVAR, METHOD)  \
  if (!i)							\
    {								\
      x = [url METHOD];						\
      i = x && IVAR ? [(id) IVAR compare: x] : (IVAR != x);	\
    }

  COMPARE (path, path);
  COMPARE (scheme ,scheme);
  COMPARE (hostname, hostname);
  COMPARE (parms, parameters);
  COMPARE (args, arguments);
  COMPARE (frag, fragment);

#undef COMPARE

  return i;
} /* -compare: */

-deepen
{
  ASGN_IVAR (path, [TLVector vectorWithSequence: path]);
  if (parms)
    ASGN_IVAR (parms, [TLVector vectorWithSequence: parms]);
  if (args)
    ASGN_IVAR (args, [TLVector vectorWithSequence: args]);

  return self;
} /* -deepen */

-(BOOL) explicitPortP
{
  return port != -1;
} /* -explicitPortP */

-(id <TLString>) filename
{
  if (!filename)
    {
      TLMutableString *s = [TLMutableString mutableString];
      id <TLEnumerator> e = [path enumerator];
      id n;

      e = [path enumerator];
      while ((n = [e nextObject]) || [e notEndP])
	formac (s, @"/%@", n ? n : @"");

      /* No length sort-of implies the root directory.  */
      if (![s length])
	[s appendCChar: '/'];

      /* XXX Immutalize.  */
      filename = s;
    }
  return filename;  
} /* -filename */

-(TLVector *) gatherFrom: (char) sc intermediate: (char) ic
  from: (const char *) s with: (int *) bp length: (int) l
{
  TLVector *v = nil;
  int b = *bp, i;

  if (b < l && s[b] == sc)
    {
      v = [TLVector vector];

      do
	{
	  int p = -1;

	  for (i = ++b; i < l && !SEPARATOR3 (s[i]) && s[i] != ic; i++)
	    if (s[i] == '=' && p == -1)
	      p = i;

	  if (p == -1)
	    [v addElement: [TLString stringWithHTMLQuotedCString: s + b
				     length: i - b]];
	  else
	    [v addElement:
		 CONS ([TLString stringWithHTMLQuotedCString: s + b
				 length: p - b],
		       [TLString stringWithHTMLQuotedCString: s + p + 1
				 length: i - p - 1])];
	  b = i;
	} while (b < l && s[b] == ic);
      *bp = b;
    }
  return (v);
} /* -gatherFrom:intermediate:from:with:length: */

-(unsigned int) hash
{
  return [(id) path hash];
} /* -hash */

-initWithString: (id <TLString>) in_name relativeTo: (id <TLURL>) base
{
  TLString *url = [in_name string];
  const char *s = [url cString];
  int b, e, i, l = [url length];

  /* Search for the scheme.  */
  for (i = 0; i < l && !SEPARATOR (s[i]); i++);
  if (i != l && s[i] == ':')
    {
      /* We got a scheme.  */
      scheme = [TLString stringWithCString: s length: i];
      b = e = i + 1;
    }
  else
    {
      /* Get the scheme from the BASE.  It defaults to NIL.  */
      scheme = base ? [base scheme] : nil;
      b = 0;
      e = i;
    }

  /* Handle the HOSTNAME and PORT.  */
  if (b == e && b < l - 1 && s[b] == '/' && s[b + 1] == '/')
    {
      /* Get the server indication.  */
      b = e = e + 2;
      for (i = e; i < l && !SEPARATOR (s[i]); i++);
      hostname = [TLString stringWithCString: s + b length: i - b];
      b = e = i;

      if (s[i] == ':')
	{
	  /* Get the explicit server port.  */
	  int all_digit = 1, value = 0;

	  for (i = b = e + 1; i < l && !SEPARATOR (s[i]); i++)
	    if (all_digit && s[i] >= '0' && s[i] <= '9')
	      value = 10 * value + s[i] - '0';
	    else
	      all_digit = 0;
	  b = e = i;

	  if (all_digit)
	    port = value;
	  else
	    {
	      /* The port indication is not numeric; try a named service.  */
	      TLString *ps = [TLString stringWithCString: s + b length: i - b];
	      TLNumber *n = [TLInetAddress portFromService: ps
					   protocol: @"tcp"];
	      if (n)
		port = [n intValue];
	      else
		INIT_RETURN_NIL ();
	    }
	}
      else
	port = -1;
    }
  else
    {
      /* Get the server and the port from the BASE.  Default to no port host
         NIL.  */
      hostname = base ? [base hostname] : nil;
      port = (base && [base explicitPortP]) ? [[base port] intValue] : -1;
    }

  /* Initialize the path vector.  */
  if (s[b] == '/')
    {
      /* An absolute path, skip the `/'.  */
      e = ++b;
      path = [TLVector vector];
    }
  else if (!base)
    {
      /* A relative path, but relative to nothing.  Don't allow that.  */
      INIT_RETURN_NIL ();
    }
  else
    {
      /* Get the relative path from the BASE.  */
      path = [TLVector vectorWithSequence: [base path]];

      /* Remove the leaf name if we're looking at a relative path, as
         relative paths concern directories; not leafs.  If we're not
         looking at a relative path (and we're thus looking at a parameter,
         query or fragment), do not discard the leaf.  */
      if (b != l && !SEPARATOR2 (s[b]))
	{
	  int len = [path length];

	  if (len)
	    [path _removeElementAtIndex: len - 1];
	}
    }

  /* Accumulate path elements.  */
  while (b < l && (s[b] == '/' || !SEPARATOR2 (s[b])))
    {
      TLString *el;

      for (i = b; i < l && !SEPARATOR2 (s[i]); i++);

      el = [TLString stringWithHTMLQuotedCString: s + b length: i - b];

      /* Handle `..' and `.' in a path now.  */
      if (![el compare: @".."])
	{
	  int len = [path length];

	  if (!len)
	    {
	      /* Won't `..' past the root!  */
	      INIT_RETURN_NIL ();
	    }
	  [path _removeElementAtIndex: len - 1];
	}
      else if ([el compare: @"."])
	[path addElement: el];

      b = i;
      if (s[b] == '/')
	{
	  b++;
	  if (b == l || (SEPARATOR2 (s[b]) && s[b] != '/'))
	    {
	      /* Add an empty last element to indicate the presence of the
                 trailing `/'.  */
	      [path addElement: nil];
	    }
	}
    }

  parms = [self gatherFrom: ';' intermediate: ';' from: s with: &b length: l];
  args = [self gatherFrom: '?' intermediate: '&' from: s with: &b length: l];

  /* Get the fragment.  */
  if (b < l && s[b] == '#')
    frag = [TLString stringWithHTMLQuotedCString: s + b + 1
		     length: l - b - 1];
  return self;
} /* -initWithString:relativeTo: */

-initWithURL: (id <TLURL>) url relativeTo: (id <TLURL>) base
{
  id o;

  scheme = [url scheme];
  hostname = [url hostname];
  path = [TLVector vectorWithSequence: [url path]];
  o = [url parameters];
  if (o)
    parms = [TLVector vectorWithSequence: o];
  o = [url arguments];
  if (o)
    args = [TLVector vectorWithSequence: o];
  frag = [url fragment];
  port = -1;
  if ([url explicitPortP])
    port = [[base port] intValue];

  if (base && !scheme)
    {
      ASGN_IVAR (scheme, [base scheme]);
      if (!hostname)
	{
	  ASGN_IVAR (hostname, [base hostname]);
	  if ([base explicitPortP] && port == -1)
	    {
	      port = [[base port] intValue];
	    }
	}
    }

  return self;
} /* -initWithURL:relativeTo: */

-(void) gcReference
{
  MARK (scheme);
  MARK (hostname);
  MARK (path);
  MARK (filename);
  MARK (parms);
  MARK (args);
  MARK (frag);
} /* -gcReference */

-(void) print: (id <TLMutableStream>) stream quoted: (BOOL) quoted
{
  if (quoted)
    formac (stream,
	    @"scheme=%#\nhost=%#\npath=%#\nparms=%#\nargs=%#\nfrag=%#\n",
	    scheme, hostname, path, parms, args, frag);
  else
    {
      if (scheme)
	formac (stream, @"%@:", scheme);
      if (hostname)
	formac (stream, @"//%@", hostname);
      if (port != -1)
	formac (stream, @":%d", port);
      [(id) [self filename] print: stream quoted: NO];
      /* XXX */
      if (parms)
	[(id) parms print: stream quoted: NO];
      /* XXX */
      if (args)
	[(id) args print: stream quoted: NO];
      if (frag)
	formac (stream, @"#%@", frag);
    }
} /* -print:quoted: */

/******************** TLURL ********************/

-(id <TLVector>) arguments
{
  return args;
} /* -arguments */

-(id <TLString>) hostname
{
  return hostname;
} /* -hostname */

-(id <TLString>) fragment
{
  return frag;
} /* -fragment */

-(id <TLVector>) parameters
{
  return parms;
} /* -parameters */

-(id <TLVector>) path
{
  return path;
} /* -path */

-(id <TLString>) scheme
{
  return scheme;
} /* -scheme */

-(id <TLNumber>) port
{
  if (port != -1)
    return [TLNumber numberWithInt: port];
  return [TLInetAddress portFromService: scheme protocol: @"tcp"];
} /* -port */

@end
