/* Implementation of TLL terminal subroutines.
   This file is part of TL, Tiggr's Library.
   Written by Tiggr <tiggr@es.ele.tue.nl>
   Copyright (C) 1995, 1996 Pieter J. Schoenmakers
   TL is distributed WITHOUT ANY WARRANTY.
   See the file LICENSE in the TL distribution for details.

   $Id: term.m,v 1.8 1996/06/07 14:03:08 tiggr Exp $  */

#import "tl/tl.h"
#import <time.h>
#import <regex.h>
#import <string.h>
#import <fcntl.h>
#import <syslog.h>
#import <sys/param.h>
#include <sys/ioctl.h>
#if HAVE_STRINGS_H
#import <strings.h>
#endif

#if !HAVE_TERMIOS
#include <sgtty.h>
#endif

/* Include <termios.h> before <sys/termios.h> to get <standards.h> on the
   NeXT.  */
#if HAVE_TERMIOS_H
#import <termios.h>
#endif
#if HAVE_SYS_TERMIOS_H
#import <sys/termios.h>
#endif

/* If there's no _POSIX_VDISABLE, it is 0 (on SunOS).  */
#ifndef _POSIX_VDISABLE
#define _POSIX_VDISABLE  0
#endif

static TLSymbol *Qignbrk, *Qbrkint, *Qignpar, *Qparmrk, *Qinpck, *Qistrip;
static TLSymbol *Qinlcr, *Qigncr, *Qicrnl, *Qiuclc, *Qixon, *Qixany, *Qixoff;
static TLSymbol *Qolcuc, *Qopost, *Qonlcr, *Qocrnl, *Qonocr, *Qonlret;
static TLSymbol *Qofill, *Qofdel, *Qnldly, *Qnl0, *Qnl1, *Qnl2, *Qnl3;
static TLSymbol *Qcrdly, *Qcr0;
static TLSymbol *Qcr1, *Qcr2, *Qcr3, *Qtabdly, *Qtab0, *Qtab1, *Qtab2;
static TLSymbol *Qtab3, *Qxtabs, *Qbsdly, *Qbs0, *Qbs1, *Qvtdly, *Qvt0, *Qvt1;
static TLSymbol *Qffdly, *Qff0, *Qff1, *Qcstopb, *Qcread, *Qparenb;
static TLSymbol *Qparodd, *Qhupcl, *Qclocal, *Qcs5, *Qcs6, *Qcs7, *Qcs8;
static TLSymbol *Qisig, *Qicanon, *Qxcase, *Qecho, *Qechoe, *Qechok;
static TLSymbol *Qechonl, *Qnoflsh, *Qtostop, *Qiexten;
static TLSymbol *Qcbreak, *Qraw, *Qpass8, *Qpass8out;
static TLSymbol *Qveof, *Qveol, *Qverase, *Qvintr, *Qvkill, *Qvmin, *Qvquit;
static TLSymbol *Qvstart, *Qvstop, *Qvsusp, *Qvswtch, *Qvtime;
static TLSymbol *Qvdsusp, *Qvrprnt, *Qvflush, *Qvwerase, *Qvlitnext;
static TLSymbol *Qispeed, *Qospeed, *Qspeed;

struct tio_symbol
{
  TLSymbol **sym;
  const char *name;
  int value;
};

#ifndef IUCLC
#define IUCLC  0
#endif

#ifndef OLCUC
#define OLCUC  0
#endif

#ifndef OCRNL
#define OCRNL  0
#endif

#ifndef ONOCR
#define ONOCR  0
#endif

#ifndef ONLRET
#define ONLRET  0
#endif

#ifndef OFILL
#define OFILL  0
#endif

#ifndef OFDEL
#define OFDEL  0
#endif

#ifndef XCASE
#define XCASE  0
#endif

#ifndef NLDLY
#define NLDLY  0
#endif

#ifndef CRDLY
#define CRDLY  0
#endif

#ifndef TABDLY
#define TABDLY  0
#endif

#ifndef BSDLY
#define BSDLY  0
#endif

#ifndef VTDLY
#define VTDLY  0
#endif

#ifndef FFDLY
#define FFDLY  0
#endif

#ifndef NL1
#define NL1  0
#endif

#ifndef CR1
#define CR1  0
#endif

#ifndef CR2
#define CR2  0
#endif

#ifndef CR3
#define CR3  0
#endif

#ifndef TAB1
#define TAB1  0
#endif

#ifndef TAB2
#define TAB2  0
#endif

#ifndef TAB3
#define TAB3  0
#endif

#ifndef BS1
#define BS1  0
#endif

#ifndef VT1
#define VT1  0
#endif

#ifndef FF1
#define FF1  0
#endif

#ifndef VSWTCH
#define VSWTCH  0
#endif

static const struct tio_symbol tio_symbols[] =
{
  {&Qignbrk, "ignbrk", -1},
  {&Qbrkint, "brkint", -1},
  {&Qignpar, "ignpar", -1},
  {&Qparmrk, "parmrk", -1},
  {&Qinpck, "inpck", -1},
  {&Qistrip, "istrip", -1},
  {&Qinlcr, "inlcr", -1},
  {&Qigncr, "igncr", -1},
  {&Qicrnl, "icrnl", -1},
  {&Qiuclc, "iuclc", -1},
  {&Qixon, "ixon", -1},
  {&Qixany, "ixany", -1},
  {&Qixoff, "ixoff", -1},
  {&Qopost, "opost", -1},
  {&Qolcuc, "olcuc", -1},
  {&Qonlcr, "onlcr", -1},
  {&Qocrnl, "ocrnl", -1},
  {&Qonocr, "onocr", -1},
  {&Qonlret, "onlret", -1},
  {&Qofill, "ofill", -1},
  {&Qofdel, "ofdel", -1},
  {&Qcstopb, "cstopb", -1},
  {&Qcread, "cread", -1},
  {&Qparenb, "parenb", -1},
  {&Qparodd, "parodd", -1},
  {&Qhupcl, "hupcl", -1},
  {&Qclocal, "clocal", -1},
  {&Qcs5, "cs5", -1},
  {&Qcs6, "cs6", -1},
  {&Qcs7, "cs7", -1},
  {&Qcs8, "cs8", -1},
  {&Qisig, "isig", -1},
  {&Qicanon, "icanon", -1},
  {&Qxcase, "xcase", -1},
  {&Qecho, "echo", -1},
  {&Qechoe, "echoe", -1},
  {&Qechok, "echok", -1},
  {&Qechonl, "echonl", -1},
  {&Qnoflsh, "noflsh", -1},
  {&Qtostop, "tostop", -1},
  {&Qiexten, "iexten", -1},
  {&Qpass8, "pass8", -1}, /* non termios.  include in termios code.  */
  {&Qpass8out, "pass8out", -1}, /* non termios.  include in termios code.  */

  {&Qnldly, "nldly", -1},
  {&Qnl0, "nl0", -1},
  {&Qnl1, "nl1", -1},
  {&Qnl2, "nl2", -1}, /* non termios.  include in termios code.  */
  {&Qnl3, "nl3", -1}, /* non termios.  include in termios code.  */
  {&Qcrdly, "crdly", -1},
  {&Qcr0, "cr0", -1},
  {&Qcr1, "cr1", -1},
  {&Qcr2, "cr2", -1},
  {&Qcr3, "cr3", -1},
  {&Qtabdly, "tabdly", -1},
  {&Qtab0, "tab0", -1},
  {&Qtab1, "tab1", -1},
  {&Qtab2, "tab2", -1},
  {&Qtab3, "tab3", -1},
  {&Qxtabs, "xtabs", -1}, /* non termios.  include in termios code.  */
  {&Qbsdly, "bsdly", -1},
  {&Qbs0, "bs0", -1},
  {&Qbs1, "bs1", -1},
  {&Qvtdly, "vtdly", -1},
  {&Qvt0, "vt0", -1},
  {&Qvt1, "vt1", -1},
  {&Qffdly, "ffdly", -1},
  {&Qff0, "ff0", -1},
  {&Qff1, "ff1", -1},

  {&Qcbreak, "cbreak", -1},
  {&Qraw, "raw", -1},

  {&Qveof, "veof", VEOF},
  {&Qveol, "veol", VEOL},
  {&Qverase, "verase", VERASE},
  {&Qvintr, "vintr", VINTR},
  {&Qvkill, "vkill", VKILL},
  {&Qvmin, "vmin", VMIN},
  {&Qvquit, "vquit", VQUIT},
  {&Qvstart, "vstart", VSTART},
  {&Qvstop, "vstop", VSTOP},
  {&Qvsusp, "vsusp", VSUSP},
  {&Qvswtch, "vswtch", VSWTCH},
  {&Qvtime, "vtime", VTIME},

  {&Qvdsusp, "vdsusp", -1}, /* non termios.  include in termios code.  */
  {&Qvrprnt, "vrprnt", -1}, /* non termios.  include in termios code.  */
  {&Qvflush, "vflush", -1}, /* non termios.  include in termios code.  */
  {&Qvwerase, "vwerase", -1}, /* non termios.  include in termios code.  */
  {&Qvlitnext, "vlitnext", -1}, /* non termios.  include in termios code.  */

  {&Qispeed, "ispeed", -1},
  {&Qospeed, "ospeed", -1},
  {&Qspeed, "speed", -1},

  {0, 0}
};

/* Define the subroutines in this file.  */
static struct tl_subdef subdefs[] =
{
  DEFSUB (nil, @"ctermid",	Fctermid,	0, 0),
  DEFSUB (nil, @"tty-lock",	Ftty_lock,	1, 2),
  DEFSUB (nil, @"tty-unlock",	Ftty_unlock,	1, 1),
  DEFSUB (nil, @"ttyname",	Fttyname,	1, 1),
  DEFSUB (nil, @"tcgetattr",	Ftcgetattr,	1, 1),
  DEFSUB (nil, @"tcsetattr",	Ftcsetattr,	2, 2),
  DEFSUB (nil, @"tcgetpgrp",	Ftcgetpgrp,	0, 1),
  DEFSUB (nil, @"tcsetpgrp",	Ftcsetpgrp,	0, 2),
  DEFSUB (0, 0, 0, 0, 0)
};

#if HAVE_TERMIOS
/* Symbols from the control characters.  */
static TLSymbol **tio_ccs[] =
{
  &Qvmin, &Qvtime,

/* All symbols up to this index are printed as numbers.  Starting at
   this index, they're printed as characters.  */
#define CCS_FIRST_INDEX_STRING	2

  &Qveof, &Qveol, &Qverase, &Qvintr, &Qvkill, &Qvquit,
  &Qvstart, &Qvstop, &Qvsusp, &Qvswtch, 0
};
#else /* not HAVE_TERMIOS */
static TLSymbol **tio_ccs[] =
{
  /* sgttyb */
  &Qverase, &Qvkill,
  /* tchars */
  &Qvintr, &Qvquit, &Qvstart, &Qvstop, &Qveof, &Qveol,
  /* ltchars */
  &Qvsusp, &Qvdsusp, &Qvrprnt, &Qvflush, &Qvwerase, &Qvlitnext,
  0
};
#endif

/* Mapping from baudrate to something one could actually call a baud rate.  */
static struct baudrate_map
{
    unsigned int baudrate;
    int code;
} tio_baud[] =
{
#ifdef B0
    {0, B0},
#endif
#ifdef B50
    {50, B50},
#endif
#ifdef B75
    {75, B75},
#endif
#ifdef B110
    {110, B110},
#endif
#ifdef B134
    {134, B134},
#endif
#ifdef B150
    {150, B150},
#endif
#ifdef B200
    {200, B200},
#endif
#ifdef B300
    {300, B300},
#endif
#ifdef B600
    {600, B600},
#endif
#ifdef B900
    {900, B900},
#endif
#ifdef B1200
    {1200, B1200},
#endif
#ifdef B1800
    {1800, B1800},
#endif
#ifdef B2400
    {2400, B2400},
#endif
#ifdef B3600
    {3600, B3600},
#endif
#ifdef B4800
    {4800, B4800},
#endif
#ifdef B7200
    {7200, B7200},
#endif
#ifdef B9600
    {9600, B9600},
#endif
#ifdef B14400
    {14400, B14400},
#endif
#ifdef B19200
    {19200, B19200},
#endif
#ifdef B28800
    {28800, B28800},
#endif
#ifdef B38400
    {38400, B38400},
#endif
#ifdef B43200
    {43200, B43200},
#endif
#ifdef B57600
    {57600, B57600},
#endif
#ifdef B115200
    {115200, B115200},
#endif
#ifdef B230400
    {230400, B230400},
#endif
#ifdef B460800
    {460800, B460800},
#endif
    {0, -1},
};

static int term_did_init = 0;

static void
term_init (void)
{
  id list, last;
  int i;

  if (term_did_init)
      return;
  term_did_init = 1;

  for (i = 0; tio_symbols[i].sym; i++)
    *tio_symbols[i].sym
      = (tio_symbols[i].value == -1
	 ? [TLSymbol symbolWithName:
	    [TLString stringWithCString: tio_symbols[i].name]]
	 : [TLSymbol symbolWithName:
	    [TLString stringWithCString: tio_symbols[i].name]
	   constantValue: [TLNumber numberWithInt: tio_symbols[i].value]]);

  /* Initialize the `baud-rate-mappings'.  */
  for (i = 0, list = last = nil; tio_baud[i].code != -1; i++)
      {
	  id o = CONS ([TLNumber numberWithUnsignedInt: tio_baud[i].baudrate],
		       [TLNumber numberWithInt: tio_baud[i].code]);
	  o = CONS (o, nil);
	  if (last)
	      [last rplacd: o];
	  else
	      list = o;
	  last = o;
      }
  [[TLSymbol symbolWithName: @"baud-rate-mappings"] setVarValue: list];
} /* term_init */

static unsigned int
baud_from_code (int code)
{
    int i;

    for (i = 0; tio_baud[i].code != -1; i++)
	if (tio_baud[i].code == code)
	    return (tio_baud[i].baudrate);
    return (0);
} /* baud_from_code */

void
tll_term_init (void)
{
  [TLLSubroutine registerSubroutines: subdefs];
} /* tll_term_init */

id
Fctermid (void)
{
#if HAVE_CTERMID
  return ([TLString stringWithCString: ctermid (0)]);
#else
  /* XXX Should check whether we have a ctty!  */
  return (@"/dev/tty");
#endif
} /* Fctermid */

id
Ftty_lock (id <TLString> name)
{
  char buf[MAXPATHLEN + 1];

  sprintf (buf, TTY_LOCK_FILENAME, [name cString]);
  /* XXX Should return the pid of the locking process instead of plain NIL.  */
  return (fsdo_lock (buf, 1, NULL) ? Qt : nil);
} /* Ftty_lock */

id
Ftty_unlock (id <TLString> name)
{
  char buf[MAXPATHLEN + 1];

  sprintf (buf, TTY_LOCK_FILENAME, [name cString]);
  /* XXX Should return the pid of the locking process instead of plain NIL.  */
  return (fsdo_unlock (buf, 1) ? Qt : nil);
} /* Ftty_unlock */

id
Fttyname (id <TLStream> s)
{
  int fd = [s fileDescriptor];
  const char *n = ttyname (fd);

  return (n ? [TLString stringWithCString: n] : nil);
} /* Fttyname */

/******************** termios ********************/

id
Ftcgetpgrp (id <TLStream> stream)
{
  int i, pg, fd = [(stream ? stream : V_stdout_) fileDescriptor];

#if HAVE_TERMIOS
  i = pg = tcgetpgrp (fd);
#else
  i = ioctl (fd, TIOCGPGRP, &pg);
#endif

  if (i >= 0)
    return ([TLNumber numberWithInt: pg]);
  formac (V_stderr_, @"tcgetpgrp (%# (%d), %#): %s\n", stream, fd, ERRMSG);
  return (nil);
} /* Ftcgetpgrp */

id
Ftcsetpgrp (id <TLStream> stream, id <TLNumber> pgid)
{
  int fd = [(stream ? stream : V_stdout_) fileDescriptor];
  int i, pg = [(pgid ? pgid : Fgetpgrp (nil)) intValue];

#ifdef HAVE_TERMIOS
  i = tcsetpgrp (fd, pg);
#else
  i = ioctl (fd, TIOCSPGRP, &pg);
#endif

  if (i >= 0)
    return (Qt);
  formac (V_stderr_, @"tcsetpgrp (%# (%d), %# (%d)): %s\n",
	  stream, fd, pgid, pg, ERRMSG);
  return (nil);
} /* Ftcsetpgrp */

id
Ftcgetattr (id <TLStream> stream)
{
  int fd = [stream fileDescriptor];
  TLCons *flags = nil;

#if HAVE_TERMIOS
  struct termios tio;
  int i;

  term_init ();

  if (tcgetattr (fd, &tio))
    [TLLSubroutine error: "tcgetattr %#: %s", stream, ERRMSG];

  if (IGNBRK && (tio.c_iflag & IGNBRK)) flags = CONS (Qignbrk, flags);
  if (BRKINT && (tio.c_iflag & BRKINT)) flags = CONS (Qbrkint, flags);
  if (IGNPAR && (tio.c_iflag & IGNPAR)) flags = CONS (Qignpar, flags);
  if (PARMRK && (tio.c_iflag & PARMRK)) flags = CONS (Qparmrk, flags);
  if (INPCK && (tio.c_iflag & INPCK)) flags = CONS (Qinpck, flags);
  if (ISTRIP && (tio.c_iflag & ISTRIP)) flags = CONS (Qistrip, flags);
  if (INLCR && (tio.c_iflag & INLCR)) flags = CONS (Qinlcr, flags);
  if (IGNCR && (tio.c_iflag & IGNCR)) flags = CONS (Qigncr, flags);
  if (ICRNL && (tio.c_iflag & ICRNL)) flags = CONS (Qicrnl, flags);
  if (IUCLC && (tio.c_iflag & IUCLC)) flags = CONS (Qiuclc, flags);
  if (IXON && (tio.c_iflag & IXON)) flags = CONS (Qixon, flags);  
  if (IXANY && (tio.c_iflag & IXANY)) flags = CONS (Qixany, flags);
  if (IXOFF && (tio.c_iflag & IXOFF)) flags = CONS (Qixoff, flags);
  if (OPOST && (tio.c_oflag & OPOST)) flags = CONS (Qopost, flags);
  if (OLCUC && (tio.c_oflag & OLCUC)) flags = CONS (Qolcuc, flags);
  if (ONLCR && (tio.c_oflag & ONLCR)) flags = CONS (Qonlcr, flags);
  if (OCRNL && (tio.c_oflag & OCRNL)) flags = CONS (Qocrnl, flags);
  if (ONOCR && (tio.c_oflag & ONOCR)) flags = CONS (Qonocr, flags);
  if (ONLRET && (tio.c_oflag & ONLRET)) flags = CONS (Qonlret, flags);
  if (OFILL && (tio.c_oflag & OFILL)) flags = CONS (Qofill, flags);
  if (OFDEL && (tio.c_oflag & OFDEL)) flags = CONS (Qofdel, flags);
  if (CSTOPB && (tio.c_cflag & CSTOPB)) flags = CONS (Qcstopb, flags);
  if (CREAD && (tio.c_cflag & CREAD)) flags = CONS (Qcread, flags);
  if (PARENB && (tio.c_cflag & PARENB)) flags = CONS (Qparenb, flags);
  if (PARODD && (tio.c_cflag & PARODD)) flags = CONS (Qparodd, flags);
  if (HUPCL && (tio.c_cflag & HUPCL)) flags = CONS (Qhupcl, flags);
  if (CLOCAL && (tio.c_cflag & CLOCAL)) flags = CONS (Qclocal, flags);
  if (ISIG && (tio.c_lflag & ISIG)) flags = CONS (Qisig, flags);
  if (ICANON && (tio.c_lflag & ICANON)) flags = CONS (Qicanon, flags);
  if (XCASE && (tio.c_lflag & XCASE)) flags = CONS (Qxcase, flags);
  if (ECHO && (tio.c_lflag & ECHO)) flags = CONS (Qecho, flags);
  if (ECHOE && (tio.c_lflag & ECHOE)) flags = CONS (Qechoe, flags);
  if (ECHOK && (tio.c_lflag & ECHOK)) flags = CONS (Qechok, flags);
  if (ECHONL && (tio.c_lflag & ECHONL)) flags = CONS (Qechonl, flags);
  if (NOFLSH && (tio.c_lflag & NOFLSH)) flags = CONS (Qnoflsh, flags);
  if (TOSTOP && (tio.c_lflag & TOSTOP)) flags = CONS (Qtostop, flags);
  if (IEXTEN && (tio.c_lflag & IEXTEN)) flags = CONS (Qiexten, flags);

  if ((tio.c_cflag & CSIZE) == CS5) flags = CONS (Qcs5, flags);
  else if ((tio.c_cflag & CSIZE) == CS6) flags = CONS (Qcs6, flags);
  else if ((tio.c_cflag & CSIZE) == CS7) flags = CONS (Qcs7, flags);
  else if ((tio.c_cflag & CSIZE) == CS8) flags = CONS (Qcs8, flags);

  /* Don't show zero delays.  */
#if NLDLY
  if ((tio.c_oflag & NLDLY) == NL1) flags = CONS (Qnl1, flags);
#endif
#if CRDLY
  if (tio.c_oflag & CRDLY)
    if ((tio.c_oflag & CRDLY) == CR1) flags = CONS (Qcr1, flags);
    else if ((tio.c_oflag & CRDLY) == CR2) flags = CONS (Qcr2, flags);
    else if ((tio.c_oflag & CRDLY) == CR3) flags = CONS (Qcr3, flags);
#endif
#if TABDLY
  if (tio.c_oflag & TABDLY)
    if ((tio.c_oflag & TABDLY) == TAB1) flags = CONS (Qtab1, flags);
    else if ((tio.c_oflag & TABDLY) == TAB2) flags = CONS (Qtab2, flags);
    else if ((tio.c_oflag & TABDLY) == TAB3) flags = CONS (Qtab3, flags);
#endif
#if BSDLY
  if ((tio.c_oflag & BSDLY) == BS1) flags = CONS (Qbs1, flags);
#endif
#if VTDLY
  if ((tio.c_oflag & VTDLY) == VT1) flags = CONS (Qvt1, flags);
#endif
#if FFDLY
  if ((tio.c_oflag & FFDLY) == FF1) flags = CONS (Qff1, flags);
#endif

  /* Put the points last, to please the eye.  */
  {
    TLCons *list = nil;
    speed_t is, os;

    for (i = 0; tio_ccs[i]; i++)
      {
	int idx = [[*tio_ccs[i] varValue] intValue];

	if (tio.c_cc[idx] != _POSIX_VDISABLE)
	  list = CONS (CONS (*tio_ccs[i],
			     ((i < CCS_FIRST_INDEX_STRING)
			      ? [TLNumber numberWithInt: tio.c_cc[idx]]
			      : [TLString stringWithCString:
				 &tio.c_cc[idx] length: 1])), list);
      }

    is = cfgetispeed (&tio);
    os = cfgetospeed (&tio);

    list = CONS (CONS (Qispeed, [TLNumber numberWithInt: baud_from_code (is)]),
		 CONS (CONS (Qospeed, [TLNumber numberWithInt:
				       baud_from_code (os)]), list));
    if (flags)
	[flags nconc: list];
    else
	flags = list;
  }
#else /* not HAVE_TERMIOS  */
  struct sgttyb sg;
  struct tchars tc;
  struct ltchars lc;
  int lm, discipline;

  char *tc_chars[] =
    {
      /* sg */
      &sg.sg_erase, &sg.sg_kill,
      /* tchars */
      &tc.t_intrc, &tc.t_quitc, &tc.t_startc,
      &tc.t_stopc, &tc.t_eofc, &tc.t_brkc,
      /* ltchars */
      &lc.t_suspc, &lc.t_dsuspc, &lc.t_rprntc,
      &lc.t_flushc, &lc.t_werasc, &lc.t_lnextc,
      0
    };

  term_init ();

  /* Force the new tty discipline.  */
  if (ioctl (fd, TIOCGETD, &discipline))
    [TLLSubroutine error: "ioctl TIOCGETD %#: %s", stream, ERRMSG];
  if (discipline == OTTYDISC)
      {
	discipline = NTTYDISC;
	if (ioctl (fd, TIOCSETD, &discipline))
	    [TLLSubroutine error: "ioctl TIOCSETD %#: %s", stream, ERRMSG];
      }

  if (ioctl (fd, TIOCGETP, &sg) || ioctl (fd, TIOCGETC, &tc)
      || ioctl (fd, TIOCGLTC, &lc) || ioctl (fd, TIOCLGET, &lm))
    [TLLSubroutine error: "ioctl %#: %s", stream, ERRMSG];

  if (lm & LLITOUT)
      sg.sg_flags &= ~(CRMOD | ECHO);
  if (sg.sg_flags & (CRMOD | ECHO)) flags = CONS (Qopost, flags);
  if (sg.sg_flags & (LCASE)) flags = CONS (Qiexten, flags);

  if (sg.sg_flags & (EVENP | ODDP))
      {
	flags = CONS (Qparenb, flags);
	if (sg.sg_flags & ODDP)
	    flags = CONS (Qparodd, flags);
      }
  if (!(sg.sg_flags & (RAW | CBREAK)))
      flags = CONS (Qicanon, flags);
  if (sg.sg_flags & CRMOD)
      flags = CONS (Qinlcr, CONS (Qonlcr, flags));
  if (sg.sg_flags & LCASE)
      flags = CONS (Qiuclc, CONS (Qolcuc, flags));

  /* Don't show zero delays.  */
  if (sg.sg_flags & NLDELAY)
    if ((sg.sg_flags & NLDELAY) == NL1) flags = CONS (Qnl1, flags);
    else if ((sg.sg_flags & NLDELAY) == NL2) flags = CONS (Qnl2, flags);
    else if ((sg.sg_flags & NLDELAY) == NL3) flags = CONS (Qnl3, flags);
  if (sg.sg_flags & CRDELAY)
    if ((sg.sg_flags & CRDELAY) == CR1) flags = CONS (Qcr1, flags);
    else if ((sg.sg_flags & CRDELAY) == CR2) flags = CONS (Qcr2, flags);
    else if ((sg.sg_flags & CRDELAY) == CR3) flags = CONS (Qcr3, flags);
  if (sg.sg_flags & TBDELAY)
    if ((sg.sg_flags & TBDELAY) == TAB1) flags = CONS (Qtab1, flags);
    else if ((sg.sg_flags & TBDELAY) == TAB2) flags = CONS (Qtab2, flags);
    else if ((sg.sg_flags & TBDELAY) == XTABS) flags = CONS (Qxtabs, flags);
  if ((sg.sg_flags & BSDELAY) == BS1) flags = CONS (Qbs1, flags);
  if ((sg.sg_flags & VTDELAY) == FF1) flags = CONS (Qvt1, flags);

  {
    TLCons *list = nil;
    int i;

    for (i = 0; tc_chars[i]; i++)
	if (*tc_chars[i] != -1)
	    list = CONS (CONS (*tio_ccs[i],
			       [TLString stringWithCString: tc_chars[i]
				length: 1]), list);

    list = CONS (CONS (Qispeed, [TLNumber numberWithInt:
				 baud_from_code (sg.sg_ispeed)]),
		 CONS (CONS (Qospeed, [TLNumber numberWithInt:
				       baud_from_code (sg.sg_ospeed)]), nil));
    if (flags)
	[flags nconc: list];
    else
	flags = list;
  }
#endif /* not HAVE_TERMIOS  */

  return (flags);
} /* Ftcgetattr */

id
Ftcsetattr (id <TLStream> stream, id flags)
{
  int i, fd = [stream fileDescriptor];
  id arg;

#if HAVE_TERMIOS
  struct termios tio;

  term_init ();

  bzero ((void *) &tio, sizeof (tio));
  for (i = 0; i < NCCS; i++)
    tio.c_cc[i] = _POSIX_VDISABLE;

  while (flags)
    {
      DECONS (flags, arg, flags);

      if (arg == Qignbrk) tio.c_iflag |= IGNBRK;
      else if (arg == Qbrkint) tio.c_iflag |= BRKINT;
      else if (arg == Qignpar) tio.c_iflag |= IGNPAR;
      else if (arg == Qparmrk) tio.c_iflag |= PARMRK;
      else if (arg == Qinpck) tio.c_iflag |= INPCK;
      else if (arg == Qistrip) tio.c_iflag |= ISTRIP;
      else if (arg == Qinlcr) tio.c_iflag |= INLCR;
      else if (arg == Qigncr) tio.c_iflag |= IGNCR;
      else if (arg == Qicrnl) tio.c_iflag |= ICRNL;
      else if (arg == Qiuclc) tio.c_iflag |= IUCLC;
      else if (arg == Qixon) tio.c_iflag |= IXON;
      else if (arg == Qixany) tio.c_iflag |= IXANY;
      else if (arg == Qixoff) tio.c_iflag |= IXOFF;
      else if (arg == Qopost) tio.c_oflag |= OPOST;
      else if (arg == Qolcuc) tio.c_oflag |= OLCUC;
      else if (arg == Qonlcr) tio.c_oflag |= ONLCR;
      else if (arg == Qocrnl) tio.c_oflag |= OCRNL;
      else if (arg == Qonocr) tio.c_oflag |= ONOCR;
      else if (arg == Qonlret) tio.c_oflag |= ONLRET;
      else if (arg == Qofill) tio.c_oflag |= OFILL;
      else if (arg == Qofdel) tio.c_oflag |= OFDEL;
      else if (arg == Qcstopb) tio.c_cflag |= CSTOPB;
      else if (arg == Qcread) tio.c_cflag |= CREAD;
      else if (arg == Qparenb) tio.c_cflag |= PARENB;
      else if (arg == Qparodd) tio.c_cflag |= PARODD;
      else if (arg == Qhupcl) tio.c_cflag |= HUPCL;
      else if (arg == Qclocal) tio.c_cflag |= CLOCAL;
      else if (arg == Qisig) tio.c_lflag |= ISIG;
      else if (arg == Qicanon) tio.c_lflag |= ICANON;
      else if (arg == Qxcase) tio.c_lflag |= XCASE;
      else if (arg == Qecho) tio.c_lflag |= ECHO;
      else if (arg == Qechoe) tio.c_lflag |= ECHOE;
      else if (arg == Qechok) tio.c_lflag |= ECHOK;
      else if (arg == Qechonl) tio.c_lflag |= ECHONL;
      else if (arg == Qnoflsh) tio.c_lflag |= NOFLSH;
      else if (arg == Qtostop) tio.c_lflag |= TOSTOP;
      else if (arg == Qiexten) tio.c_lflag |= IEXTEN;
      else if (arg == Qcs5) tio.c_cflag = ((tio.c_cflag) & ~CSIZE) | CS5;
      else if (arg == Qcs6) tio.c_cflag = ((tio.c_cflag) & ~CSIZE) | CS6;
      else if (arg == Qcs7) tio.c_cflag = ((tio.c_cflag) & ~CSIZE) | CS7;
      else if (arg == Qcs8) tio.c_cflag = ((tio.c_cflag) & ~CSIZE) | CS8;
      else if (arg == Qnl0 || arg == Qcr0 || arg == Qtab0 || arg == Qbs0
	       || arg == Qvt0 || arg == Qff0);
      else if (arg == Qnl1) tio.c_oflag = (tio.c_oflag & ~NLDLY) | NL1;
      else if (arg == Qcr1) tio.c_oflag = (tio.c_oflag & ~CRDLY) | CR1;
      else if (arg == Qcr2) tio.c_oflag = (tio.c_oflag & ~CRDLY) | CR2;
      else if (arg == Qcr3) tio.c_oflag = (tio.c_oflag & ~CRDLY) | CR3;
      else if (arg == Qtab1) tio.c_oflag = (tio.c_oflag & ~TABDLY) | TAB1;
      else if (arg == Qtab2) tio.c_oflag = (tio.c_oflag & ~TABDLY) | TAB2;
      else if (arg == Qtab3) tio.c_oflag = (tio.c_oflag & ~TABDLY) | TAB3;
      else if (arg == Qbs1) tio.c_oflag = (tio.c_oflag & ~BSDLY) | BS1;
      else if (arg == Qvt1) tio.c_oflag = (tio.c_oflag & ~VTDLY) | VT1;
      else if (arg == Qff1) tio.c_oflag = (tio.c_oflag & ~FFDLY) | FF1;
      else if (!CONSP (arg))
	[TLLSubroutine warning: "%#: bad tcattribute flag", arg];
      else
	{
	  id <TLString, TLNumber> value;
	  int v = 0;
	  id key;

	  DECONS (arg, key, value);
	  for (i = 0; tio_ccs[i]; i++)
	    if (key == *tio_ccs[i])
	      break;
	  if (tio_ccs[i])
	    {
	      if (STRINGP (value))
		if ([value length] == 1)
		  v = [[value string] characterAtIndex: 0];
		else
		  {
		    [TLLSubroutine warning: "%#: bad tcattribute value", arg];
		    v = -1;
		  }
	      else if ([(id) value conformsTo: @protocol (TLNumber)])
		v = [value intValue];

	      if (v != -1)
		tio.c_cc[[[*tio_ccs[i] varValue] intValue]] = v;
	    }
	  else if (key == Qospeed || key == Qispeed || key == Qspeed)
	    {
	      TLSymbol *s = [TLSymbol symbolWithName: @"baud-rate-mappings"];
	      TLCons *m = [s varValue], *map;
	      int r = 0;

	      /* XXX assoc!  */
	      while (m)
		{
		  DECONS (m, map, m);
		  if ([[map car] equal: value])
		    break;
		}

	      if (!m)
		[TLLSubroutine error: "%#: bad baud-rate", arg];

	      if (key == Qospeed || key == Qspeed)
		r = r || cfsetospeed (&tio, [[map cdr] intValue]);
	      if (key == Qispeed || key == Qspeed)
		r = r || cfsetispeed (&tio, [[map cdr] intValue]);
	      if (r)
		[TLLSubroutine error: "%#: bad baud-rate (%s)", arg, ERRMSG];
	    }
	  else
	    [TLLSubroutine error: "%#: bad tcattribute", arg];
	}
    }

  if (tcsetattr (fd, TCSANOW, &tio))
    [TLLSubroutine error: "tcsetattr %#: %s", stream, ERRMSG];

#else /* not HAVE_TERMIOS */
  struct sgttyb sg;
  struct tchars tc;
  struct ltchars lc;
  int lm = 0;

  char *tc_chars[] =
    {
      /* sg */
      &sg.sg_erase, &sg.sg_kill,
      /* tchars */
      &tc.t_intrc, &tc.t_quitc, &tc.t_startc,
      &tc.t_stopc, &tc.t_eofc, &tc.t_brkc,
      /* ltchars */
      &lc.t_suspc, &lc.t_dsuspc, &lc.t_rprntc,
      &lc.t_flushc, &lc.t_werasc, &lc.t_lnextc,
      0
    };

  term_init ();

  bzero (&sg, sizeof (sg));
  bzero (&tc, sizeof (tc));
  bzero (&lc, sizeof (lc));

  for (i = 0; tc_chars[i]; i++)
      *tc_chars[i] = -1;

  while (flags)
    {
      DECONS (flags, arg, flags);

      if (arg == Qcbreak) sg.sg_flags |= CBREAK;
      else if (arg == Qraw) sg.sg_flags |= RAW;
      else if (arg == Qnl0 || arg == Qcr0 || arg == Qtab0 || arg == Qbs0
	       || arg == Qvt0 || arg == Qff0);
      else if (arg == Qnl1) sg.sg_flags = (sg.sg_flags & ~NLDELAY) | NL1;
      else if (arg == Qcr1) sg.sg_flags = (sg.sg_flags & ~CRDELAY) | CR1;
      else if (arg == Qcr2) sg.sg_flags = (sg.sg_flags & ~CRDELAY) | CR2;
      else if (arg == Qcr3) sg.sg_flags = (sg.sg_flags & ~CRDELAY) | CR3;
      else if (arg == Qtab1) sg.sg_flags = (sg.sg_flags & ~TBDELAY) | TAB1;
      else if (arg == Qtab2 || arg == Qtab3)
	  sg.sg_flags = (sg.sg_flags & ~TBDELAY) | TAB2;
      else if (arg == Qxtabs) sg.sg_flags = (sg.sg_flags & ~TBDELAY) | XTABS;
      else if (arg == Qbs1) sg.sg_flags = (sg.sg_flags & ~BSDELAY) | BS1;
      else if (arg == Qvt1 || arg == Qff1)
	  sg.sg_flags = (sg.sg_flags & ~VTDELAY) | FF1;
      else if (arg == Qpass8) lm |= LPASS8;
      else if (arg == Qpass8out) lm |= LPASS8OUT;
      else if (!CONSP (arg))
	[TLLSubroutine warning: "%#: bad tcattribute flag", arg];
      else
	{
	  id <TLString, TLNumber> value;
	  int v = 0;
	  id key;

	  DECONS (arg, key, value);
	  for (i = 0; tio_ccs[i]; i++)
	    if (key == *tio_ccs[i])
	      break;
	  if (tio_ccs[i])
	    {
	      if (STRINGP (value))
		if ([value length] == 1)
		  v = [[value string] characterAtIndex: 0];
		else
		  {
		    [TLLSubroutine warning: "%#: bad tcattribute value", arg];
		    v = -1;
		  }
	      else if ([(id) value conformsTo: @protocol (TLNumber)])
		v = [value intValue];

	      if (v != -1)
		  *tc_chars[i] = v;
	    }
	  else if (key == Qospeed || key == Qispeed || key == Qspeed)
	    {
	      TLSymbol *s = [TLSymbol symbolWithName: @"baud-rate-mappings"];
	      TLCons *m = [s varValue], *map;

	      /* XXX assoc!  */
	      while (m)
		{
		  DECONS (m, map, m);
		  if ([[map car] equal: value])
		    break;
		}

	      if (!m)
		[TLLSubroutine warning: "%#: bad baud-rate", arg];
	      else
		{
		  if (key == Qospeed || key == Qspeed)
		    sg.sg_ospeed = [[map cdr] intValue];
		  if (key == Qispeed || key == Qspeed)
		    sg.sg_ispeed = [[map cdr] intValue];
		}
	    }
	  else
	    [TLLSubroutine warning: "%#: bad tcattribute", arg];
	}
    }

  if (ioctl (fd, TIOCSETN, &sg) || ioctl (fd, TIOCSETC, &tc)
      || ioctl (fd, TIOCSLTC, &lc) || ioctl (fd, TIOCLSET, &lm))
    [TLLSubroutine error: "ioctl %#: %s", stream, ERRMSG];

#endif

  return (stream);
} /* Ftcsetattr */
